@@ -1,4 +1,4 @@
-#! /usr/bin/env python2.4
+#! /usr/bin/env python
 
 # mklibs.py: An automated way to create a minimal /lib/ directory.
 #
@@ -40,7 +40,6 @@
 # TODO
 # * complete argument parsing as given as comment in main
 
-import commands
 import string
 import re
 import sys
@@ -62,13 +61,15 @@ def debug(level, *msg):
 # return a list of lines of output of the command
 def command(command, *args):
     debug(DEBUG_SPAM, "calling", command, string.join(args))
-    (status, output) = commands.getstatusoutput(command + ' ' + string.join(args))
-    if os.WEXITSTATUS(status) != 0:
+    pipe = os.popen(command + ' ' + ' '.join(args), 'r')
+    output = pipe.read().strip()
+    status = pipe.close() 
+    if status is not None and os.WEXITSTATUS(status) != 0:
         print "Command failed with status", os.WEXITSTATUS(status),  ":", \
                command, string.join(args)
 	print "With output:", output
         sys.exit(1)
-    return string.split(output, '\n')
+    return [i for i in output.split('\n') if i]
 
 # Filter a list according to a regexp containing a () group. Return
 # a set.
@@ -82,36 +83,41 @@ def regexpfilter(list, regexp, groupnr = 1):
 
     return result
 
+def elf_header(obj):
+    if not os.access(obj, os.F_OK):
+        raise "Cannot find lib: " + obj
+    output = command("mklibs-readelf", "--print-elf-header", obj)
+    s = [int(i) for i in output[0].split()]
+    return {'class': s[0], 'data': s[1], 'machine': s[2], 'flags': s[3]}
+
 # Return a set of rpath strings for the passed object
 def rpath(obj):
     if not os.access(obj, os.F_OK):
         raise "Cannot find lib: " + obj
-    output = command(target + "objdump", "--private-headers", obj)
-    return [root + "/" + x for x in regexpfilter(output, ".*RPATH\s*(\S+)$")]
+    output = command("mklibs-readelf", "--print-rpath", obj)
+    return [root + "/" + x for x in output]
 
 # Return a set of libraries the passed objects depend on.
 def library_depends(obj):
     if not os.access(obj, os.F_OK):
         raise "Cannot find lib: " + obj
-    output = command(target + "objdump", "--private-headers", obj)
-    return regexpfilter(output, ".*NEEDED\s*(\S+)$")
+    return command("mklibs-readelf", "--print-needed", obj)
 
 # Return a list of libraries the passed objects depend on. The
 # libraries are in "-lfoo" format suitable for passing to gcc.
 def library_depends_gcc_libnames(obj):
     if not os.access(obj, os.F_OK):
         raise "Cannot find lib: " + obj
-    output = command(target + "objdump", "--private-headers", obj)
-    output = regexpfilter(output, ".*NEEDED\s*lib(\S+)\.so.*$")
-    return ' '.join(["-l%s" % x for x in output])
-
-# Scan readelf output. Example:
-# Num Buc:    Value  Size Type  Bind  Vis     Ndx Name
-#   1   1: 080497e4   168 FUNC GLOBAL DEFAULT UND strchr
-# or in some cases on alpha:
-#  2007 335: 0000000000084d00   168    FUNC GLOBAL DEFAULT [<other>: 80]   10 strchr
-symline_regexp = \
-    re.compile("\s*\d+\s+\d+: .+\s+\d+\s+\w+\s+(\w+)+\s+\w+\s+(?:\[<other>: \d+\]\s+)*(\w+)\s+([^\s]+)$")
+    libs = library_depends(obj)
+    ret = []
+    for i in libs:
+        match = re.match("^(((?P<ld>ld\S*)|(lib(?P<lib>\S+))))\.so.*$", i)
+        if match:
+            if match.group('ld'):
+                ret.append(find_lib(match.group(0)))
+            elif match.group('lib'):
+                ret.append('-l%s' % match.group('lib'))
+    return ' '.join(ret)
 
 # Return undefined symbols in an object as a set of tuples (name, weakness)
 def undefined_symbols(obj):
@@ -119,14 +125,14 @@ def undefined_symbols(obj):
         raise "Cannot find lib" + obj
 
     result = set()
-    output = command(target + "readelf", "-s", "-D", "-W", obj)
+    output = command("mklibs-readelf", "--print-symbols-undefined", obj)
     for line in output:
-        match = symline_regexp.match(line)
-        if match:
-            bind, ndx, name = match.groups()
-            if bind != "LOCAL" and ndx == "UND":
-                debug(DEBUG_SPAM, "undefined_symbols adding", `name`)
-                result.add((name, bind == "WEAK"))
+        name, weak_string = line.split(' ', 2)
+        weak = False
+        if weak_string == 'True':
+            weak = True
+        debug(DEBUG_SPAM, "undefined_symbols adding: %s, %s" % (name, weak))
+        result.add((name, weak))
     return result
 
 # Return a set of symbols provided by a library
@@ -135,15 +141,11 @@ def provided_symbols(obj):
         raise "Cannot find lib" + obj
 
     result = set()
-    debug(DEBUG_SPAM, "provided_symbols result = ", `result`)
-    output = command(target + "readelf", "-s", "-D", "-W", obj)
+    output = command("mklibs-readelf", "--print-symbols-provided", obj)
     for line in output:
-        match = symline_regexp.match(line)
-        if match:
-            bind, ndx, name = match.groups()
-            if bind != "LOCAL" and not ndx in ("UND", "ABS"):
-                debug(DEBUG_SPAM, "provided_symbols adding ", `name`)
-                result.add(name)
+        name, weak = line.split(' ', 2)
+        debug(DEBUG_SPAM, "provided_symbols adding %s, %s" %(name, weak))
+        result.add(name)
     return result
     
 # Return real target of a symlink
@@ -185,12 +187,11 @@ def find_pic_map(lib):
     return ""
 
 def extract_soname(so_file):
-    soname_data = regexpfilter(command(target + "readelf", "--all", "-W", so_file),
-                               ".*SONAME.*\[(.*)\].*")
+    soname_data = command("mklibs-readelf", "--print-soname", so_file)
     if soname_data:
         return soname_data.pop()
-
     return ""
+
 def usage(was_err):
     if was_err:
         outfd = sys.stderr
@@ -307,14 +308,10 @@ for prog in proglist:
         objects[inode] = prog
 
 if not ldlib:
-    pattern = re.compile(".*Requesting program interpreter:.*/([^\]/]+).*")
     for obj in objects.values():
-        output = command(target + "readelf", "--program-headers", obj)
-	for x in output:
-	    match = pattern.match(x)
-	    if match:
-	        ldlib = match.group(1)
-	        break
+        output = command("mklibs-readelf", "--print-interp", obj)
+        if output:
+            ldlib = output.pop()
 	if ldlib:
 	    break
 
@@ -373,16 +370,11 @@ while 1:
         needed_symbols.update(undefined_symbols(obj))
         libraries.update(library_depends(obj))
 
-    # FIXME: on i386 this is undefined but not marked UND
-    # I don't know how to detect those symbols but this seems
-    # to be the only one and including it on alpha as well
-    # doesn't hurt. I guess all archs can live with this.
-    needed_symbols.add(("sys_siglist", 1))
-
     # calculate what symbols are present in small_libs and available_libs
     present_symbols = set()
     checked_libs = small_libs
     checked_libs.extend(available_libs)
+    checked_libs.append(ldlib)
     for lib in checked_libs:
         present_symbols.update(provided_symbols(lib))
 
@@ -415,6 +407,11 @@ while 1:
     library_symbols_used = {}
     symbol_provider = {}
 
+    # WORKAROUND: Always add libgcc on old-abi arm
+    header = elf_header(find_lib(libraries.copy().pop()))
+    if header['machine'] == 40 and header['flags'] & 0xff000000 == 0:
+        libraries.add('libgcc_s.so.1')
+
     # Calculate all symbols each library provides
     for library in libraries:
         path = find_lib(library)
@@ -475,38 +472,39 @@ while 1:
                 continue
             debug(DEBUG_SPAM, "soname:", soname)
 
-            if library_symbols_used[library]:
-                symbols = library_symbols_used[library]
-            else:
-                symbols = []
+            symbols = set()
+            extra_flags = []
+            extra_pre_obj = []
+            extra_post_obj = []
+
+            symbols.update(library_symbols_used[library])
 
             # libc.so.6 needs its soinit.o and sofini.o as well as the pic
-            if (soname == "libc.so.6" or soname == "libc.so.6.1"):
+            if soname in ("libc.so.6", "libc.so.6.1"):
                 # force dso_handle.os to be included, otherwise reduced libc
                 # may segfault in ptmalloc_init due to undefined weak reference
-                extra_flags = find_lib(ldlib)
-                extra_pre_obj = libc_extras_dir + "/soinit.o"
-                extra_post_obj = libc_extras_dir + "/sofini.o"
+                extra_pre_obj.append(libc_extras_dir + "/soinit.o")
+                extra_post_obj.append(libc_extras_dir + "/sofini.o")
                 symbols.add('__dso_handle')
-            else:
-                extra_flags = ""
-                extra_pre_obj = ""
-                extra_post_obj = ""
+
             map_file = find_pic_map(library)
             if map_file:
-                extra_flags = extra_flags + " -Wl,--version-script=" + map_file
+                extra_flags.append("-Wl,--version-script=" + map_file)
+
             # compile in only used symbols
-            command(target + "gcc",
-                "-nostdlib -nostartfiles -shared -Wl,-soname=" + soname,\
-                ' '.join(["-u%s" % a for a in symbols]), \
-                "-o", dest_path + "/" + so_file_name + "-so", \
-                extra_pre_obj, \
-                pic_file, \
-                extra_post_obj, \
-                extra_flags, \
-                "-lgcc", \
-                ' '.join(["-L%s" % a for a in [dest_path] + lib_path]), \
-                library_depends_gcc_libnames(so_file))
+            cmd = []
+            cmd.append("-nostdlib -nostartfiles -shared -Wl,-soname=" + soname)
+            cmd.extend(["-u%s" % a for a in symbols])
+            cmd.extend(["-o", dest_path + "/" + so_file_name + "-so"])
+            cmd.extend(extra_pre_obj)
+            cmd.append(pic_file)
+            cmd.extend(extra_post_obj)
+            cmd.extend(extra_flags)
+            cmd.append("-lgcc")
+            cmd.extend(["-L%s" % a for a in [dest_path] + lib_path])
+            cmd.append(library_depends_gcc_libnames(so_file))
+            command(target + "gcc", *cmd)
+
             # strip result
             command(target + "objcopy", "--strip-unneeded -R .note -R .comment",
                       dest_path + "/" + so_file_name + "-so",
@@ -537,8 +535,8 @@ for lib in regexpfilter(os.listdir(dest_path), "(.*so[.\d]*)$"):
         os.rename(dest_path + "/" + lib, dest_path + "/" + soname)
 
 # Make sure the dynamic linker is present and is executable
-ld_file = find_lib(ldlib)
-ld_file_name = os.path.basename(ld_file)
+ld_file_name = os.path.basename(ldlib)
+ld_file = find_lib(ld_file_name)
 
 if not os.access(dest_path + "/" + ld_file_name, os.F_OK):
     debug(DEBUG_NORMAL, "I: stripping and copying dynamic linker.")
